import pickle

import numpy as np
import tensorflow as tf

from chia import instrumentation, knowledge
from chia.components.classifiers.keras_hierarchicalclassification import (
    EmbeddingBasedKerasHC,
)


class OneHotEmbeddingBasedKerasHC(EmbeddingBasedKerasHC, instrumentation.Observable):
    def __init__(self, kb, l2=5e-5):
        instrumentation.Observable.__init__(self)
        EmbeddingBasedKerasHC.__init__(self, kb)

        # Configuration
        self._l2_regularization_coefficient = l2

        self.last_observed_concept_count = len(
            self.kb.concepts(flags={knowledge.ConceptFlag.PREDICTION_TARGET})
        )

        self.fc_layer = None
        self.uid_to_dimension = {}
        self.dimension_to_uid = []

        self.update_embedding()

    def update_embedding(self):
        current_observed_concepts = self.kb.concepts(
            flags={knowledge.ConceptFlag.PREDICTION_TARGET}
        )
        current_observed_concept_count = len(current_observed_concepts)

        try:
            old_weights = self.fc_layer.get_weights()
        except Exception:
            old_weights = []

        self.fc_layer = tf.keras.layers.Dense(
            current_observed_concept_count,
            activation="softmax",
            kernel_regularizer=tf.keras.regularizers.l2(
                self._l2_regularization_coefficient
            )
            if self._l2_regularization_coefficient > 0.0
            else None,
        )

        update_uids = [
            concept.uid
            for concept in current_observed_concepts
            if concept.uid not in self.dimension_to_uid
        ]

        self.dimension_to_uid += sorted(update_uids)

        self.uid_to_dimension = {
            uid: dimension for dimension, uid in enumerate(self.dimension_to_uid)
        }

        if len(old_weights) == 2:
            # Layer can be updated
            new_weights = np.concatenate(
                [
                    old_weights[0],
                    np.zeros(
                        [
                            old_weights[0].shape[0],
                            current_observed_concept_count
                            - self.last_observed_concept_count,
                        ]
                    ),
                ],
                axis=1,
            )
            new_biases = np.concatenate(
                [
                    old_weights[1],
                    np.zeros(
                        current_observed_concept_count
                        - self.last_observed_concept_count
                    ),
                ],
                axis=0,
            )

            self.fc_layer.build([None, old_weights[0].shape[0]])

            self.fc_layer.set_weights([new_weights, new_biases])
        self.report_metric("observed_concepts", current_observed_concept_count)
        self.last_observed_concept_count = current_observed_concept_count

    def predict_embedded(self, feature_batch):
        return self.fc_layer(feature_batch)

    def embed(self, labels):
        embeddings = []
        for label in labels:
            if label == "chia::UNCERTAIN":
                embeddings += [
                    np.full(
                        self.last_observed_concept_count,
                        fill_value=1.0 / self.last_observed_concept_count,
                    )
                ]
            else:
                embeddings += [
                    tf.one_hot(
                        self.uid_to_dimension[label],
                        depth=self.last_observed_concept_count,
                    )
                ]

        return tf.stack(embeddings)

    def deembed_dist(self, embedded_labels):
        return [
            [(uid, embedded_label[dim]) for uid, dim in self.uid_to_dimension.items()]
            for embedded_label in embedded_labels
        ]

    def loss(self, feature_batch, ground_truth):
        embedded_predictions = self.predict_embedded(feature_batch)
        embedded_ground_truth = self.embed(ground_truth)
        loss = tf.reduce_mean(
            tf.keras.losses.categorical_crossentropy(
                embedded_ground_truth, embedded_predictions
            )
        )
        return loss

    def observe(self, samples, gt_resource_id):
        self.maybe_update_embedding()

    def regularization_losses(self):
        return self.fc_layer.losses

    def trainable_variables(self):
        return self.fc_layer.trainable_variables

    def save(self, path):
        with open(path + "_hc.pkl", "wb") as target:
            pickle.dump(self.fc_layer.get_weights(), target)

        with open(path + "_uidtodim.pkl", "wb") as target:
            pickle.dump((self.uid_to_dimension, self.dimension_to_uid), target)

    def restore(self, path):
        with open(path + "_hc.pkl", "rb") as target:
            new_weights = pickle.load(target)
            has_weights = False
            try:
                has_weights = len(self.fc_layer.get_weights()) == 2
            except Exception:
                pass

            if not has_weights:
                self.fc_layer.build([None, new_weights[0].shape[0]])

            self.fc_layer.set_weights(new_weights)

        with open(path + "_uidtodim.pkl", "rb") as target:
            (self.uid_to_dimension, self.dimension_to_uid) = pickle.load(target)

        self.update_embedding()



from chia import components
from chia.components.runners import epoch
from chia.components.runners.runner import Runner


class RunnerFactory(components.Factory):
    name_to_class_mapping = {"epoch": epoch.EpochRunner}


__all__ = ["Runner", "RunnerFactory"]



import abc

from chia import instrumentation


class Runner(abc.ABC, instrumentation.Observable):
    def __init__(self, experiment_container):
        instrumentation.Observable.__init__(self)
        self.experiment_container = experiment_container

    @abc.abstractmethod
    def run(self):
        pass



import networkx as nx
from networkx.algorithms import shortest_paths

from chia import knowledge
from chia.components.evaluators.evaluator import Evaluator


class HierarchicalEvaluator(Evaluator):
    def __init__(self, kb: knowledge.KnowledgeBase):
        self.kb = kb
        self.sample_count = 0
        self.running_cm = dict()

    def reset(self):
        self.sample_count = 0
        self.running_cm = dict()

    def update(self, samples, gt_resource_id, prediction_resource_id):
        for sample in iter(samples):
            gt_uid = sample.get_resource(gt_resource_id)
            pred_uid = sample.get_resource(prediction_resource_id)

            # Confusion Matrix
            if (gt_uid, pred_uid) in self.running_cm.keys():
                self.running_cm[(gt_uid, pred_uid)] += 1
            else:
                self.running_cm[(gt_uid, pred_uid)] = 1

            self.sample_count += 1

    def result(self):
        rgraph = self.kb.get_hyponymy_relation_rgraph()
        root = next(nx.topological_sort(rgraph))

        # Semantic Distance
        running_distance = 0
        running_correct_depth = 0
        running_error_depth = 0
        running_hierarchical_precision = 0
        running_hierarchical_recall = 0

        sample_count_confused = 0
        ugraph = self.kb.get_hyponymy_relation().ugraph()
        for (gt_uid, pred_uid), count in self.running_cm.items():
            lca = nx.lowest_common_ancestor(rgraph, gt_uid, pred_uid)
            lca_depth = nx.shortest_path_length(rgraph, root, lca) + 1

            if gt_uid != pred_uid:
                running_distance += shortest_paths.shortest_path_length(
                    ugraph, gt_uid, pred_uid
                )
                sample_count_confused += count

                # Error Depth in Hierarchy

                running_error_depth += lca_depth * count

            running_correct_depth += lca_depth * count

            # Hierarchical Precision / Recall
            gt_ancestors = set(nx.ancestors(rgraph, gt_uid)) | {gt_uid}
            pred_ancestors = set(nx.ancestors(rgraph, pred_uid)) | {pred_uid}

            hierarchical_precision = len(
                pred_ancestors.intersection(gt_ancestors)
            ) / float(len(pred_ancestors))
            hierarchical_recall = len(
                pred_ancestors.intersection(gt_ancestors)
            ) / float(len(gt_ancestors))

            running_hierarchical_precision += hierarchical_precision * count
            running_hierarchical_recall += hierarchical_recall * count

        # Cannot use tuples as keys in json...
        nested_cm = dict()
        for (gt_uid, pred_uid), count in self.running_cm.items():
            if gt_uid not in nested_cm.keys():
                nested_cm[gt_uid] = dict()

            nested_cm[gt_uid][pred_uid] = count

        ret_dict = dict()
        if self.sample_count > 0:
            ret_dict.update(
                {
                    "semantic_distance": float(running_distance)
                    / float(self.sample_count),
                    "abs_confusion_matrix": nested_cm,
                    "correct_depth": float(running_correct_depth)
                    / float(self.sample_count),
                    "hierarchical_recall": float(running_hierarchical_recall)
                    / float(self.sample_count),
                    "hierarchical_precision": float(running_hierarchical_precision)
                    / float(self.sample_count),
                    "accuracy_from_cm": float(self.sample_count - sample_count_confused)
                    / float(self.sample_count),
                }
            )
        if sample_count_confused > 0:
            ret_dict.update(
                {
                    "semantic_distance_confused": float(running_distance)
                    / float(sample_count_confused),
                    "error_depth": float(running_error_depth)
                    / float(sample_count_confused),
                }
            )

        return ret_dict



import time

from PIL import Image

from chia import instrumentation


class NetworkResistantImage:
    @staticmethod
    def open(filename: str, observable: instrumentation.Observable):
        cumulative_wait_max = 7200.0
        wait_interval_initial = 0.5

        # Set initial state
        wait_interval = wait_interval_initial
        cumulative_wait = 0.0
        last_exception = None

        while cumulative_wait <= cumulative_wait_max:
            try:
                image = Image.open(filename)
                return image

            except Exception as ex:
                observable.log_warning(
                    f"Cannot open {filename}. Waiting {wait_interval} seconds."
                )
                last_exception = ex
                time.sleep(wait_interval)
                cumulative_wait += wait_interval
                wait_interval *= 2.0

        raise last_exception



import random

import numpy as np

from chia import data
from chia.components.datasets import dataset
from chia.knowledge import relation

_namespace_uid = "iCIFAR"

_label_names = [
    "apple",
    "aquarium_fish",
    "baby",
    "bear",
    "beaver",
    "bed",
    "bee",
    "beetle",
    "bicycle",
    "bottle",
    "bowl",
    "boy",
    "bridge",
    "bus",
    "butterfly",
    "camel",
    "can",
    "castle",
    "caterpillar",
    "cattle",
    "chair",
    "chimpanzee",
    "clock",
    "cloud",
    "cockroach",
    "couch",
    "crab",
    "crocodile",
    "cup",
    "dinosaur",
    "dolphin",
    "elephant",
    "flatfish",
    "forest",
    "fox",
    "girl",
    "hamster",
    "house",
    "kangaroo",
    "keyboard",
    "lamp",
    "lawn_mower",
    "leopard",
    "lion",
    "lizard",
    "lobster",
    "man",
    "maple_tree",
    "motorcycle",
    "mountain",
    "mouse",
    "mushroom",
    "oak_tree",
    "orange",
    "orchid",
    "otter",
    "palm_tree",
    "pear",
    "pickup_truck",
    "pine_tree",
    "plain",
    "plate",
    "poppy",
    "porcupine",
    "possum",
    "rabbit",
    "raccoon",
    "ray",
    "road",
    "rocket",
    "rose",
    "sea",
    "seal",
    "shark",
    "shrew",
    "skunk",
    "skyscraper",
    "snail",
    "snake",
    "spider",
    "squirrel",
    "streetcar",
    "sunflower",
    "sweet_pepper",
    "table",
    "tank",
    "telephone",
    "television",
    "tiger",
    "tractor",
    "train",
    "trout",
    "tulip",
    "turtle",
    "wardrobe",
    "whale",
    "willow_tree",
    "wolf",
    "woman",
    "worm",
]

_labels_to_wordnet = [
    ("apple", "apple.n.01"),
    ("aquarium_fish", "freshwater_fish.n.01"),
    ("baby", "baby.n.01"),
    ("bear", "bear.n.01"),
    ("beaver", "beaver.n.07"),
    ("bed", "bed.n.01"),
    ("bee", "bee.n.01"),
    ("beetle", "beetle.n.01"),
    ("bicycle", "bicycle.n.01"),
    ("bottle", "bottle.n.01"),
    ("bowl", "bowl.n.01"),
    ("boy", "male_child.n.01"),
    ("bridge", "bridge.n.01"),
    ("bus", "bus.n.01"),
    ("butterfly", "butterfly.n.01"),
    ("camel", "camel.n.01"),
    ("can", "can.n.01"),
    ("castle", "castle.n.02"),
    ("caterpillar", "caterpillar.n.01"),
    ("cattle", "cattle.n.01"),
    ("chair", "chair.n.01"),
    ("chimpanzee", "chimpanzee.n.01"),
    ("clock", "clock.n.01"),
    ("cloud", "cloud.n.02"),
    ("cockroach", "cockroach.n.01"),
    ("couch", "sofa.n.01"),
    ("crab", "crab.n.01"),
    ("crocodile", "crocodile.n.01"),
    ("cup", "cup.n.01"),
    ("dinosaur", "dinosaur.n.01"),
    ("dolphin", "dolphin.n.02"),
    ("elephant", "elephant.n.01"),
    ("flatfish", "flatfish.n.02"),
    ("forest", "forest.n.01"),
    ("fox", "fox.n.01"),
    ("girl", "female_child.n.01"),
    ("hamster", "hamster.n.01"),
    ("house", "house.n.01"),
    ("kangaroo", "kangaroo.n.01"),
    ("keyboard", "computer_keyboard.n.01"),
    ("lamp", "lamp.n.01"),
    ("lawn_mower", "lawn_mower.n.01"),
    ("leopard", "leopard.n.02"),
    ("lion", "lion.n.01"),
    ("lizard", "lizard.n.01"),
    ("lobster", "lobster.n.02"),
    ("man", "man.n.01"),
    ("maple_tree", "maple.n.02"),
    ("motorcycle", "motorcycle.n.01"),
    ("mountain", "mountain.n.01"),
    ("mouse", "mouse.n.01"),
    ("mushroom", "mushroom.n.02"),
    ("oak_tree", "oak.n.02"),
    ("orange", "orange.n.01"),
    ("orchid", "orchid.n.01"),
    ("otter", "otter.n.02"),
    ("palm_tree", "palm.n.03"),
    ("pear", "pear.n.01"),
    ("pickup_truck", "pickup.n.01"),
    ("pine_tree", "pine.n.01"),
    ("plain", "plain.n.01"),
    ("plate", "plate.n.04"),
    ("poppy", "poppy.n.01"),
    ("porcupine", "porcupine.n.01"),
    ("possum", "opossum.n.02"),
    ("rabbit", "rabbit.n.01"),
    ("raccoon", "raccoon.n.02"),
    ("ray", "ray.n.07"),
    ("road", "road.n.01"),
    ("rocket", "rocket.n.01"),
    ("rose", "rose.n.01"),
    ("sea", "sea.n.01"),
    ("seal", "seal.n.09"),
    ("shark", "shark.n.01"),
    ("shrew", "shrew.n.02"),
    ("skunk", "skunk.n.04"),
    ("skyscraper", "skyscraper.n.01"),
    ("snail", "snail.n.01"),
    ("snake", "snake.n.01"),
    ("spider", "spider.n.01"),
    ("squirrel", "squirrel.n.01"),
    ("streetcar", "streetcar.n.01"),
    ("sunflower", "sunflower.n.01"),
    ("sweet_pepper", "sweet_pepper.n.02"),
    ("table", "table.n.02"),
    ("tank", "tank.n.01"),
    ("telephone", "telephone.n.01"),
    ("television", "television.n.02"),
    ("tiger", "tiger.n.02"),
    ("tractor", "tractor.n.01"),
    ("train", "train.n.01"),
    ("trout", "trout.n.02"),
    ("tulip", "tulip.n.01"),
    ("turtle", "turtle.n.02"),
    ("wardrobe", "wardrobe.n.01"),
    ("whale", "whale.n.02"),
    ("willow_tree", "willow.n.01"),
    ("wolf", "wolf.n.01"),
    ("woman", "woman.n.01"),
    ("worm", "worm.n.01"),
]


class iCIFARDataset(dataset.Dataset):
    def __init__(self):
        import tensorflow as tf

        (
            (self.train_X, self.train_y),
            (self.test_X, self.test_y),
        ) = tf.keras.datasets.cifar100.load_data(label_mode="fine")

        self.sequence_seed = 19219
        self._update_sequence()

        sorting_sequence_train = np.argsort(self.train_y[:, 0], kind="stable")
        self.train_X = self.train_X[sorting_sequence_train]
        self.train_y = self.train_y[sorting_sequence_train]

        # Attributes are assigned in setup()
        self.classes_per_batch = None
        self.setup()

    def setup(self, classes_per_batch=4, **kwargs):
        self.classes_per_batch = classes_per_batch

    def train_pool_count(self):
        return self.get_train_pool_count(self.classes_per_batch)

    def test_pool_count(self):
        return 1

    def train_pool(self, index, label_resource_id):
        return self.get_train_pool_for(index, label_resource_id, self.classes_per_batch)

    def test_pool(self, index, label_resource_id):
        assert index == 0
        return self.get_test_pool(label_resource_id)

    def namespace(self):
        return _namespace_uid

    def _update_sequence(self):
        random.seed(self.sequence_seed)
        self.sequence = random.sample(range(100), 100)

    def get_train_pool_count(self, classes_per_batch):
        return 100 // classes_per_batch

    def get_train_pool_for(self, batch, label_resource_id, classes_per_batch):
        assert (100 % classes_per_batch) == 0
        batch_count = 100 // classes_per_batch
        assert batch < batch_count

        samples = []
        classes_for_pool = self.sequence[
            (batch * classes_per_batch) : (batch + 1) * classes_per_batch
        ]

        # print(f"Retrieving images for classes {classes_for_pool}")
        for class_ in classes_for_pool:
            training_data_range = list(range(500 * class_, 500 * (class_ + 1)))
            class_X = self.train_X[500 * class_ : 500 * (class_ + 1)]
            class_y = self.train_y[500 * class_ : 500 * (class_ + 1)]
            samples += self._build_samples(
                class_X, class_y, training_data_range, label_resource_id, "train"
            )

        return samples

    def get_test_pool(self, label_resource_id):
        return self._build_samples(
            self.test_X, self.test_y, range(0, 10000), label_resource_id, "test"
        )

    def _build_samples(self, X, y, data_range, label_resource_id, prefix):
        assert X.shape[0] == len(data_range)
        samples = []
        for i, data_id in enumerate(data_range):
            class_label = y[i, 0]
            np_image = X[i]
            samples += [
                data.Sample(
                    source=self.__class__.__name__,
                    uid=f"{_namespace_uid}::{prefix}.{data_id}",
                )
                .add_resource(self.__class__.__name__, "input_img_np", np_image)
                .add_resource(
                    self.__class__.__name__,
                    label_resource_id,
                    f"{_namespace_uid}::{_label_names[int(class_label)]}",
                )
            ]
        return samples

    def get_hyponymy_relation_source(self):
        return relation.StaticRelationSource(
            [
                (f"{_namespace_uid}::{label}", f"WordNet3.0::{synset}")
                for label, synset in _labels_to_wordnet
            ]
        )

    def prediction_targets(self):
        return [
            f"{_namespace_uid}::{concept_uid}" for concept_uid, _ in _labels_to_wordnet
        ]



import os

import numpy as np
from PIL import Image

from chia import data, instrumentation
from chia.components.datasets import dataset
from chia.helpers import NetworkResistantImage
from chia.knowledge import relation

_namespace_uid = "NABirds"


class NABirdsDataset(dataset.Dataset, instrumentation.Observable):
    def __init__(self, base_path, side_length, use_lazy_mode=True):
        instrumentation.Observable.__init__(self)
        self.base_path = base_path
        self.side_length = side_length
        self.use_lazy_mode = use_lazy_mode

        with open(os.path.join(self.base_path, "classes.txt")) as cls:
            lines = [x.strip() for x in cls]
            tuples = [x.split(sep=" ", maxsplit=1) for x in lines]
            tuples = [(int(k), str(v)) for (k, v) in tuples]

            self.uid_for_label_id = {
                k: f"{_namespace_uid}::{int(k):03d}{v}" for (k, v) in tuples
            }

            self.nabirds_ids = {k for (k, v) in tuples}

            if len([k for (k, v) in tuples]) != len({k for (k, v) in tuples}):
                raise ValueError("Non-unique IDs found!")

        with open(os.path.join(self.base_path, "image_class_labels.txt")) as lab:
            with open(os.path.join(self.base_path, "train_test_split.txt")) as tts:
                with open(os.path.join(self.base_path, "images.txt")) as iid:
                    lablines = [x.strip() for x in lab]
                    labtuples = [x.split(sep=" ", maxsplit=1) for x in lablines]
                    labtuples = [
                        (str(some_primary_key), int(label))
                        for (some_primary_key, label) in labtuples
                    ]

                    ttslines = [x.strip() for x in tts]
                    ttstuples = [x.split(sep=" ", maxsplit=1) for x in ttslines]
                    ttstuples = [
                        (str(some_primary_key), int(is_train_or_test))
                        for (some_primary_key, is_train_or_test) in ttstuples
                    ]

                    iidlines = [x.strip() for x in iid]
                    iidtuples = [x.split(sep=" ", maxsplit=1) for x in iidlines]
                    iidtuples = [
                        (str(some_primary_key), str(image_path))
                        for (some_primary_key, image_path) in iidtuples
                    ]
                    self.image_location_for_image_id = {k: v for (k, v) in iidtuples}

                    combinedtuples = [a + b for (a, b) in zip(labtuples, ttstuples)]
                    mismatches = [a != c for (a, b, c, d) in combinedtuples]
                    if any(mismatches):
                        raise ValueError("Mismatch between tts and label files!")

                    combinedtuples = [(a, b, d) for (a, b, c, d) in combinedtuples]

                    self._nabirds_training_tuples = [
                        (img, id) for (img, id, tt) in combinedtuples if tt == 1
                    ]
                    self._nabirds_validation_tuples = [
                        (img, id) for (img, id, tt) in combinedtuples if tt == 0
                    ]

        with open(os.path.join(self.base_path, "hierarchy.txt")) as hie:
            lines = [x.strip() for x in hie]
            tuples = [x.split(sep=" ", maxsplit=1) for x in lines]
            self.tuples = [
                (self.uid_for_label_id[int(k)], self.uid_for_label_id[int(v)])
                for (k, v) in tuples
            ]

    def setup(self, **kwargs):
        pass

    def train_pool_count(self):
        return 1

    def test_pool_count(self):
        return 1

    def train_pool(self, index, label_resource_id):
        assert index == 0
        return self.get_train_pool(label_resource_id)

    def test_pool(self, index, label_resource_id):
        assert index == 0
        return self.get_test_pool(label_resource_id)

    def namespace(self):
        return _namespace_uid

    def get_train_pool(self, label_resource_id):
        return [
            self._build_sample(image_id, label_id, label_resource_id, "train")
            for image_id, label_id in self._nabirds_training_tuples
        ]

    def get_test_pool(self, label_resource_id):
        return [
            self._build_sample(image_id, label_id, label_resource_id, "test")
            for image_id, label_id in self._nabirds_validation_tuples
        ]

    def _build_sample(self, image_id, label_id, label_resource_id, split):
        sample_ = data.Sample(
            source=self.__class__.__name__, uid=f"{_namespace_uid}::{split}:{image_id}"
        ).add_resource(
            self.__class__.__name__, label_resource_id, self.uid_for_label_id[label_id],
        )
        if self.use_lazy_mode:
            sample_ = sample_.add_resource(
                self.__class__.__name__,
                "image_location",
                os.path.join(
                    self.base_path, "images", self.image_location_for_image_id[image_id]
                ),
            ).add_lazy_resource(
                self.__class__.__name__, "input_img_np", self._load_from_location
            )
        else:
            sample_ = sample_.add_resource(
                self.__class__.__name__,
                "image_location",
                os.path.join(
                    self.base_path, "images", self.image_location_for_image_id[image_id]
                ),
            )
            sample_ = sample_.add_resource(
                self.__class__.__name__,
                "input_img_np",
                self._load_from_location(sample_),
            )
        return sample_

    def _load_from_location(self, sample_):
        im = NetworkResistantImage.open(
            sample_.get_resource("image_location"), self
        ).resize((self.side_length, self.side_length), Image.ANTIALIAS)
        if im.mode != "RGB":
            im = im.convert("RGB")
        return np.asarray(im)

    def get_hyponymy_relation_source(self):
        return relation.StaticRelationSource(self.tuples)

    def prediction_targets(self):
        label_ids = {label_id for _, label_id in self._nabirds_training_tuples} | {
            label_id for _, label_id in self._nabirds_validation_tuples
        }
        return {self.uid_for_label_id[label_id] for label_id in label_ids}



import json
import os

import numpy as np
from PIL import Image

from chia import data, instrumentation
from chia.components.datasets import dataset
from chia.helpers import NetworkResistantImage
from chia.knowledge import relation

_namespace_uid = "iNaturalist2018"


class iNaturalist2018Dataset(dataset.Dataset, instrumentation.Observable):
    def setup(self, **kwargs):
        pass

    def train_pool_count(self):
        return 1

    def test_pool_count(self):
        return 1

    def train_pool(self, index, label_resource_id):
        assert index == 0
        return self.pool_from_json("train2018.json", label_resource_id)

    def test_pool(self, index, label_resource_id):
        assert index == 0
        return self.pool_from_json("val2018.json", label_resource_id)

    def namespace(self):
        return _namespace_uid

    def __init__(self, base_path, side_length):
        instrumentation.Observable.__init__(self)
        self.base_path = base_path
        self.side_length = side_length

        self._id_to_class = dict()
        self._hypernymy_relation = set()

        def is_a(concept, superconcept):
            relation_element = (
                f"{_namespace_uid}::{concept}",
                f"{_namespace_uid}::{superconcept}",
            )
            self._hypernymy_relation |= {relation_element}

        with open(os.path.join(self.base_path, "categories.json")) as json_file:
            json_data = json.load(json_file)
            for json_datum in json_data:
                self._id_to_class[
                    json_datum["id"]
                ] = f"{_namespace_uid}::{json_datum['name']}"

                # Biology!
                is_a(f"Ki{json_datum['kingdom']}", "Root")
                is_a(f"Ph{json_datum['phylum']}", f"Ki{json_datum['kingdom']}")
                is_a(f"Cl{json_datum['class']}", f"Ph{json_datum['phylum']}")
                is_a(f"Or{json_datum['order']}", f"Cl{json_datum['class']}")
                is_a(f"Fa{json_datum['family']}", f"Or{json_datum['order']}")
                is_a(f"Ge{json_datum['genus']}", f"Fa{json_datum['family']}")
                is_a(f"{json_datum['name']}", f"Ge{json_datum['genus']}")

    def pool_from_json(self, filename, label_resource_id):
        with open(os.path.join(self.base_path, filename)) as json_file:
            json_data = json.load(json_file)
        image_list = json_data["images"]
        annotation_list = json_data["annotations"]
        annotations = {ann["image_id"]: ann["category_id"] for ann in annotation_list}

        return [
            self.build_sample(image_dict, label_resource_id, annotations)
            for image_dict in image_list
        ]

    def build_sample(self, image_dict, label_resource_id, annotations):
        image_filename = image_dict["file_name"]
        image_id = image_dict["id"]
        sample_ = data.Sample(
            source=self.__class__.__name__, uid=f"{_namespace_uid}::{image_filename}"
        )
        sample_ = sample_.add_resource(
            self.__class__.__name__,
            label_resource_id,
            self._id_to_class[annotations[image_id]],
        )
        sample_ = sample_.add_resource(
            self.__class__.__name__,
            "image_location",
            os.path.join(self.base_path, image_filename),
        ).add_lazy_resource(
            self.__class__.__name__, "input_img_np", self._load_from_location
        )

        return sample_

    def get_hyponymy_relation_source(self):
        return relation.StaticRelationSource(list(self._hypernymy_relation))

    def _load_from_location(self, sample_):
        im = NetworkResistantImage.open(
            sample_.get_resource("image_location"), self
        ).resize((self.side_length, self.side_length), Image.ANTIALIAS)
        if im.mode != "RGB":
            im = im.convert("RGB")
        return np.asarray(im)

    def prediction_targets(self):
        return set(self._id_to_class.values())



from chia.components.runners.runner import Runner


class EpochRunner(Runner):
    def __init__(self, experiment_container, epochs):
        super().__init__(experiment_container)
        self.epochs = epochs

    def run(self):
        # Get out some members of container
        dataset = self.experiment_container.dataset
        base_model = self.experiment_container.base_model

        # Build training data
        training_samples = dataset.train_pool(0, "label_gt")

        # Build test data
        test_samples = dataset.test_pool(0, "label_gt")[:100]

        # "Interact"
        training_samples = self.experiment_container.interactor.query_annotations_for(
            training_samples, "label_gt", "label_ann"
        )
        for epoch in range(self.epochs):
            self.log_info(f"Start of epoch {epoch + 1} of {self.epochs}")
            base_model.observe(training_samples, "label_ann")

            predicted_test_samples = base_model.predict(test_samples, "label_pred")

            # Go over all evaluators
            result_dict = {}
            for evaluator in self.experiment_container.evaluators:
                evaluator.update(predicted_test_samples, "label_gt", "label_pred")
                result_dict.update(evaluator.result())
                evaluator.reset()

            self.report_result(result_dict, step=epoch + 1)



class Sample:
    def __init__(self, source=None, data=None, history=None, uid=None):
        if history is not None:
            self.history = history
        elif source is not None:
            self.history = [("init", "", source)]
        else:
            self.history = []

        if data is not None:
            self.data = data
            if "_lazy_resources" not in self.data.keys():
                self.data["_lazy_resources"] = {}
        else:
            if uid is not None:
                self.data = {"uid": uid, "_lazy_resources": {}}
            else:
                raise ValueError("Need UID for sample!")

    def add_resource(self, source, resource_id, datum):
        assert resource_id not in self.data.keys()

        new_history = self.history + [("add", resource_id, source)]
        new_data = {resource_id: datum, **self.data}

        return Sample(data=new_data, history=new_history)

    def add_lazy_resource(self, source, resource_id, fn):
        new_history = self.history + [("add_lazy", resource_id, source)]

        new_lazy_resources = {resource_id: fn, **self.data["_lazy_resources"]}
        new_data = {
            "_lazy_resources": new_lazy_resources,
            **{k: v for k, v in self.data.items() if k != "_lazy_resources"},
        }

        return Sample(data=new_data, history=new_history)

    def apply_on_resource(self, source, resource_id, fn):
        assert resource_id in self.data.keys()
        new_history = self.history + [("apply", resource_id, source)]
        new_data = {k: v if k != resource_id else fn(v) for k, v in self.data.items()}

        return Sample(data=new_data, history=new_history)

    def get_resource(self, resource_id):
        if resource_id in self.data.keys():
            return self.data[resource_id]
        elif resource_id in self.data["_lazy_resources"].keys():
            return self.data["_lazy_resources"][resource_id](self)
        else:
            raise ValueError(f"Unknown resource id {resource_id}")

    def get_resource_ids(self):
        return [k for k in self.data.keys() if not k.startswith("_")] + list(
            self.data["_lazy_resources"].keys()
        )

    def has_resource(self, resource_id):
        return (
            resource_id in self.data.keys()
            or resource_id in self.data["_lazy_resources"].keys()
        )

    def __eq__(self, other):
        return self.data["uid"] == other.data["uid"]

    def __hash__(self):
        return hash(self.data["uid"])



from chia import components
from chia.instrumentation.observers import json_observer, stream_observer


class ObserverFactory(components.Factory):
    name_to_class_mapping = {
        "stream": stream_observer.StreamObserver,
        "json": json_observer.JSONObserver,
    }



import math

import tensorflow as tf
import tensorflow_addons as tfa

from chia import components


class KerasDataAugmentation:
    def __init__(
        self,
        do_random_flip_horizontal=True,
        do_random_flip_vertical=True,
        do_random_rotate=True,
        do_random_crop=True,
        random_crop_factor=0.2,
        do_random_brightness_and_contrast=True,
        random_brightness_factor=0.05,
        random_contrast_factors=(0.7, 1.3),
        do_random_hue_and_saturation=True,
        random_hue_factor=0.08,
        random_saturation_factors=(0.6, 1.6),
        do_random_scale=True,
        random_scale_factors=(0.5, 2.0),
    ):

        self.do_random_flip_horizontal = do_random_flip_horizontal
        self.do_random_flip_vertical = do_random_flip_vertical

        self.do_random_rotate = do_random_rotate

        self.do_random_crop = do_random_crop
        self.random_crop_factor = random_crop_factor

        self.do_random_brightness_and_contrast = do_random_brightness_and_contrast
        self.random_brightness_factor = random_brightness_factor
        self.random_contrast_factors = random_contrast_factors

        self.do_random_hue_and_saturation = do_random_hue_and_saturation
        self.random_hue_factor = random_hue_factor
        self.random_saturation_factors = random_saturation_factors

        self.do_random_scale = do_random_scale
        self.random_scale_factors = random_scale_factors

    @tf.function
    def process(self, sample_batch):
        sample_batch = tf.map_fn(self._process_sample, sample_batch)
        return sample_batch

    @tf.function
    def _process_sample(self, sample):
        if self.do_random_flip_horizontal:
            sample = tf.image.random_flip_left_right(sample)
        if self.do_random_flip_vertical:
            sample = tf.image.random_flip_up_down(sample)

        if self.do_random_rotate:
            sample = tfa.image.rotate(
                sample,
                angles=tf.random.uniform(shape=[], minval=0, maxval=2.0 * math.pi),
                interpolation="BILINEAR",
            )

        if self.do_random_crop or self.do_random_scale:
            sample = self._inner_random_crop_or_scale(sample)

        if self.do_random_brightness_and_contrast:
            sample = tf.image.random_brightness(sample, self.random_brightness_factor)
            sample = tf.image.random_contrast(sample, *self.random_contrast_factors)

        if self.do_random_hue_and_saturation:
            sample = tf.image.random_hue(sample, self.random_hue_factor)
            sample = tf.image.random_saturation(sample, *self.random_saturation_factors)

        return sample

    @tf.function
    def _inner_random_crop_or_scale(self, x: tf.Tensor) -> tf.Tensor:
        if self.do_random_scale:
            scale = tf.random.uniform(
                shape=[],
                minval=self.random_scale_factors[0],
                maxval=self.random_scale_factors[1],
            )
        else:
            scale = 1.0

        width = 1.0 / scale
        height = 1.0 / scale

        if self.do_random_crop:
            left_min = -self.random_crop_factor
            left_max = 1.0 + self.random_crop_factor - width
            top_min = -self.random_crop_factor
            top_max = 1.0 + self.random_crop_factor - height
        else:
            left_min = 0.0
            left_max = 1.0 - width
            top_min = 0.0
            top_max = 1.0 - height

        left = tf.random.uniform(shape=[], minval=left_min, maxval=left_max)
        top = tf.random.uniform(shape=[], minval=top_min, maxval=top_max)
        boxes = [[top, left, top + height, left + width]]

        crop_shape = (x.shape[0], x.shape[1])
        # Create different crops for an image
        crops = tf.image.crop_and_resize(
            [x], boxes=boxes, box_indices=[0], crop_size=crop_shape
        )
        # Return a random crop
        return crops[0]


class KerasDataAugmentationFactory(components.Factory):
    name_to_class_mapping = KerasDataAugmentation



from chia.components.evaluators.evaluator import Evaluator


class AccuracyEvaluator(Evaluator):
    def __init__(self, kb):
        self.correct_count = 0
        self.sample_count = 0

    def reset(self):
        self.correct_count = 0
        self.sample_count = 0

    def update(self, samples, gt_resource_id, prediction_resource_id):
        for sample in iter(samples):
            self.sample_count += 1
            if sample.get_resource(gt_resource_id) == sample.get_resource(
                prediction_resource_id
            ):
                self.correct_count += 1

    def result(self):
        return {"accuracy": float(self.correct_count) / float(self.sample_count)}



from abc import ABC, abstractmethod

import chia.knowledge.messages
from chia import instrumentation, knowledge


class KerasHierarchicalClassifier(ABC):
    @abstractmethod
    def predict_dist(self, feature_batch):
        pass

    @abstractmethod
    def loss(self, feature_batch, ground_truth):
        pass

    @abstractmethod
    def observe(self, samples, gt_resource_id):
        pass

    @abstractmethod
    def regularization_losses(self):
        pass

    @abstractmethod
    def trainable_variables(self):
        pass

    @abstractmethod
    def save(self, path):
        pass

    @abstractmethod
    def restore(self, path):
        pass


class EmbeddingBasedKerasHC(KerasHierarchicalClassifier, instrumentation.Observer, ABC):
    def __init__(self, kb: knowledge.KnowledgeBase):
        self.kb = kb
        self.kb.register(self)

        self.is_updated = False

    @abstractmethod
    def predict_embedded(self, feature_batch):
        pass

    @abstractmethod
    def embed(self, labels):
        pass

    @abstractmethod
    def deembed_dist(self, embedded_labels):
        pass

    @abstractmethod
    def update_embedding(self):
        pass

    def predict_dist(self, feature_batch):
        embedded_predictions = self.predict_embedded(feature_batch).numpy()
        return self.deembed_dist(embedded_predictions)

    def maybe_update_embedding(self):
        if not self.is_updated:
            self.is_updated = self.update_embedding()

    def update(self, message: instrumentation.Message):
        if isinstance(
            message, chia.knowledge.messages.RelationChangeMessage
        ) or isinstance(message, chia.knowledge.messages.ConceptChangeMessage):
            self.is_updated = False



from chia import components
from chia.components.evaluators import accuracy, hierarchical, topk_accuracy
from chia.components.evaluators.evaluator import Evaluator


class EvaluatorFactory(components.Factory):
    name_to_class_mapping = {
        "accuracy": accuracy.AccuracyEvaluator,
        "topk_accuracy": topk_accuracy.TopKAccuracyEvaluator,
        "hierarchical": hierarchical.HierarchicalEvaluator,
    }


__all__ = ["EvaluatorFactory", "Evaluator"]



from abc import ABC, abstractmethod


class Evaluator(ABC):
    @abstractmethod
    def update(self, samples, gt_resource_id, prediction_resource_id):
        return None

    @abstractmethod
    def result(self):
        return None

    @abstractmethod
    def reset(self):
        pass



from chia import components
from chia.components.interactors import noisy_oracle, oracle


class InteractorFactory(components.Factory):
    name_to_class_mapping = {
        "oracle": oracle.OracleInteractor,
        "noisy_oracle": noisy_oracle.NoisyOracleInteractor,
    }



import typing

from chia import components, instrumentation
from chia.knowledge.concept import Concept, ConceptFlag
from chia.knowledge.messages import ConceptChangeMessage, RelationChangeMessage
from chia.knowledge.relation import Relation, RelationFlag, RelationSource


class KnowledgeBase(instrumentation.Observable, instrumentation.Observer):
    def __init__(self):
        instrumentation.Observable.__init__(self)
        self._concepts: typing.Dict[str, Concept] = dict()
        self._relations: typing.Dict[str, Relation] = dict()

        # Build discovery object
        from chia.knowledge.discovery import Discovery

        self._discovery = Discovery(self)
        self.register(self._discovery)

    def add_relation(self, relation: Relation):
        if relation.uid in self._relations.keys():
            raise ValueError(f"Attempted to add relation {relation.uid} twice!")
        else:
            self._relations[relation.uid] = relation

            relation.register(self)

            message = RelationChangeMessage(self._sender_name())
            self.notify(message)

    def add_concepts(self, concepts: typing.List[Concept]):
        changed = False
        for concept in concepts:
            changed |= self._add_concept(concept)

        if changed:
            message = ConceptChangeMessage(self._sender_name())
            self.notify(message)

    def _add_concept(self, concept: Concept) -> bool:
        """Adds a concept to the knowledge base or merges the flags.

        Returns: True if the knowledge base has been changed in any way."""
        if concept.uid in self._concepts.keys():
            if concept.flags != self._concepts[concept.uid].flags:
                # Have to update flags
                self._concepts[concept.uid].flags |= concept.flags
                return True
            else:
                # Nothing to do
                return False
        else:
            self._concepts[concept.uid] = concept
            return True

    def concepts(
        self, flags: typing.Optional[typing.Set[ConceptFlag]] = None
    ) -> typing.List[Concept]:
        if flags is None:
            return list(self._concepts.values())
        else:
            return [
                concept
                for concept in self._concepts.values()
                if all([flag in concept.flags for flag in flags])
            ]

    def relations(
        self, flags: typing.Optional[typing.Set[RelationFlag]] = None
    ) -> typing.List[Relation]:
        if flags is None:
            return list(self._relations.values())
        else:
            return [
                relation
                for relation in self._relations.values()
                if all([flag in relation.flags for flag in flags])
            ]

    def update(self, message: instrumentation.Message):
        # Send on RelationChangeMessage from subscribed relations to own subscribers
        if isinstance(message, RelationChangeMessage):
            self.notify(message)

    # Some quick helper methods
    def add_prediction_targets(self, uids: typing.List[str]):
        concepts = [Concept(uid, {ConceptFlag.PREDICTION_TARGET}) for uid in uids]
        self.add_concepts(concepts)

    def add_hyponymy_relation(self, sources: typing.List[RelationSource]):
        relation = Relation(
            "chia::Hyponymy",
            sources=sources,
            flags={
                RelationFlag.HYPONYMY,
                RelationFlag.EXPLORE_RIGHT,
                RelationFlag.TRANSITIVE,
            },
        )
        self.add_relation(relation)

    def get_hyponymy_relation(self) -> Relation:
        applicable_relations = self.relations(flags={RelationFlag.HYPONYMY})
        if len(applicable_relations) != 1:
            raise ValueError("Cannot uniquely resolve hyponymy relation")
        else:
            return applicable_relations[0]

    def get_hyponymy_relation_rgraph(self):
        return self.get_hyponymy_relation().rgraph()


class KnowledgeBaseFactory(components.Factory):
    name_to_class_mapping = KnowledgeBase



def batches_from(samples, batch_size):
    complete_batches = []
    current_batch = []
    for sample in samples:
        current_batch.append(sample)
        if len(current_batch) == batch_size:
            complete_batches.append(current_batch)
            current_batch = []

    if len(current_batch) > 0:
        complete_batches.append(current_batch)

    return complete_batches


def batches_from_pair(Xs, ys, batch_size):
    complete_batches = []

    current_batch_X = []
    current_batch_y = []

    for (X, y) in zip(Xs, ys):
        current_batch_X.append(X)
        current_batch_y.append(y)
        if len(current_batch_X) == batch_size:
            complete_batches.append((current_batch_X, current_batch_y))
            current_batch_X = []
            current_batch_y = []

    if len(current_batch_X) > 0:
        complete_batches.append((current_batch_X, current_batch_y))

    return complete_batches



import abc
import math

from chia import components


class KerasLearningRateSchedule(abc.ABC):
    @abc.abstractmethod
    def __call__(self, step):
        pass


class ConstantKerasLearningRateSchedule(KerasLearningRateSchedule):
    def __init__(self, initial_lr):
        self.initial_lr = initial_lr

    def __call__(self, step):
        return self.initial_lr


class SGDRKerasLearningRateSchedule(KerasLearningRateSchedule):
    def __init__(
        self,
        minimum_lr: float,
        maximum_lr: float,
        T_0: int,
        T_mult: int = 2,
        maximum_lr_decay: float = 1.0,
        warmup_steps: int = 0,
        warmup_lr: float = 0.1,
    ):
        self.minimum_lr = minimum_lr
        self.maximum_lr = maximum_lr
        self.T_0 = T_0
        self.T_mult = T_mult
        self.maximum_lr_decay = maximum_lr_decay
        self.warmup_steps = warmup_steps
        self.warmup_lr = warmup_lr

    def __call__(self, step):
        if step < self.warmup_steps:
            return self.warmup_lr

        step_cur = step
        maximum_lr_cur = self.maximum_lr
        T_i = self.T_0
        while step_cur > T_i:
            step_cur -= T_i
            T_i *= self.T_mult
            maximum_lr_cur *= self.maximum_lr_decay

        T_cur = float(step_cur)
        T_i = float(T_i)

        maximum_lr_cur = max(maximum_lr_cur, self.minimum_lr)

        eta_t = self.minimum_lr + 0.5 * (maximum_lr_cur - self.minimum_lr) * (
            1.0 + math.cos(math.pi * (T_cur / T_i))
        )
        return eta_t


class ExponentialKerasLearningRateSchedule(KerasLearningRateSchedule):
    def __init__(self, initial_lr, end_lr, decay_factor, steps_per_decay):
        self.initial_lr = initial_lr
        self.end_lr = end_lr
        self.decay_factor = decay_factor
        self.steps_per_decay = steps_per_decay

    def __call__(self, step):
        current_decay_exponent = float(step) / self.steps_per_decay
        current_decay_factor = math.pow(self.decay_factor, current_decay_exponent)
        current_lr = max(self.initial_lr * current_decay_factor, self.end_lr)
        return current_lr


class KerasLearningRateScheduleFactory(components.Factory):
    name_to_class_mapping = {
        "constant": ConstantKerasLearningRateSchedule,
        "sgdr": SGDRKerasLearningRateSchedule,
        "exponential": ExponentialKerasLearningRateSchedule,
    }
    default_section = "keras_learning_rate_schedule"



from chia.components.interactors import interactor


class OracleInteractor(interactor.Interactor):
    def query_annotations_for(self, samples, gt_resource_id, ann_resource_id):
        return [
            sample.add_resource(
                self.__class__.__name__,
                ann_resource_id,
                sample.get_resource(gt_resource_id),
            )
            for sample in samples
        ]



import sys

from chia.instrumentation.message import Message
from chia.instrumentation.observers.observer import Observer


class StreamObserver(Observer):
    def __init__(self, stream=sys.stdout):
        self.stream = stream

    def update(self, message: Message):
        print(message, file=self.stream)



import typing

from chia import instrumentation, knowledge
from chia.components import base_models, classifiers


class ModelContainer:
    def __init__(
        self,
        config,
        knowledge_base: knowledge.KnowledgeBase,
        observers: typing.Iterable[instrumentation.Observer] = (),
    ):
        self.knowledge_base = knowledge_base

        self.classifier = classifiers.ClassifierFactory.create(
            config["classifier"], kb=self.knowledge_base, observers=observers
        )

        self.base_model_container = base_models.BaseModelContainerFactory.create(
            config["base_model"], classifier=self.classifier, observers=observers
        )



import tensorflow as tf
from tensorflow.keras.applications import (
    inception_resnet_v2,
    mobilenet_v2,
    nasnet,
    resnet_v2,
)

from chia import components, helpers, instrumentation


class KerasFeatureExtractor(instrumentation.Observable):
    def __init__(
        self,
        side_length,
        use_pretrained_weights="ILSVRC2012",
        architecture="ResNet50V2",
        l2=5e-5,
        trainable=False,
    ):
        instrumentation.Observable.__init__(self)

        # Weights
        self.use_pretrained_weights = use_pretrained_weights

        # Architecture
        self.architecture = architecture
        if self.architecture == "ResNet50V2":
            self.feature_extractor = resnet_v2.ResNet50V2(
                include_top=False,
                input_tensor=None,
                input_shape=None,
                pooling="avg",
                weights="imagenet"
                if self.use_pretrained_weights == "ILSVRC2012"
                else None,
            )

        elif self.architecture == "InceptionResNetV2":
            self.feature_extractor = inception_resnet_v2.InceptionResNetV2(
                include_top=False,
                input_tensor=None,
                input_shape=None,
                pooling="avg",
                weights="imagenet"
                if self.use_pretrained_weights == "ILSVRC2012"
                else None,
            )

        elif self.architecture == "MobileNetV2":
            self.side_length = side_length
            self.feature_extractor = mobilenet_v2.MobileNetV2(
                include_top=False,
                input_tensor=None,
                input_shape=(self.side_length, self.side_length, 3),
                pooling="avg",
                weights="imagenet"
                if self.use_pretrained_weights == "ILSVRC2012"
                else None,
            )

        elif self.architecture == "NASNetMobile":
            self.side_length = side_length
            self.feature_extractor = nasnet.NASNetMobile(
                include_top=False,
                input_tensor=None,
                input_shape=(self.side_length, self.side_length, 3),
                pooling="avg",
                weights="imagenet"
                if self.use_pretrained_weights == "ILSVRC2012"
                else None,
            )

        else:
            raise ValueError(f'Unknown architecture "{self.architecture}"')

        # Load other pre-trained weights if necessary
        if (
            self.use_pretrained_weights is not None
            and self.use_pretrained_weights != "ILSVRC2012"
        ):
            self.log_info(
                f"Loading alternative pretrained weights {self.use_pretrained_weights}"
            )
            self.feature_extractor.load_weights(
                helpers.maybe_expand_path(self.use_pretrained_weights, self)
            )

        # Freezing of layers
        self.trainable = trainable
        if not self.trainable:
            for layer in self.feature_extractor.layers:
                layer.trainable = False

        # L2 Regularization
        self.l2_regularization = l2
        self._add_l2_regularizers()

    def _add_l2_regularizers(self):
        if self.l2_regularization == 0:
            return

        # Add regularizer:
        # see https://jricheimer.github.io/keras/2019/02/06/keras-hack-1/
        for layer in self.feature_extractor.layers:
            if (
                isinstance(layer, tf.keras.layers.Conv2D)
                and not isinstance(layer, tf.keras.layers.DepthwiseConv2D)
            ) or isinstance(layer, tf.keras.layers.Dense):
                layer.add_loss(
                    lambda layer=layer: tf.keras.regularizers.l2(
                        self.l2_regularization
                    )(layer.kernel)
                )
            elif isinstance(layer, tf.keras.layers.DepthwiseConv2D):
                layer.add_loss(
                    lambda layer=layer: tf.keras.regularizers.l2(
                        self.l2_regularization
                    )(layer.depthwise_kernel)
                )
            if hasattr(layer, "bias_regularizer") and layer.use_bias:
                layer.add_loss(
                    lambda layer=layer: tf.keras.regularizers.l2(
                        self.l2_regularization
                    )(layer.bias)
                )

    def __call__(self, image_batch, training):
        return self.feature_extractor(image_batch, training)


class KerasFeatureExtractorFactory(components.Factory):
    name_to_class_mapping = KerasFeatureExtractor
    default_section = "keras_feature_extractor"



from distutils import util

from setuptools import find_packages, setup

main_ns = {}
ver_path = util.convert_path("chia/version.py")

with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="chia",
    version=main_ns["__version__"],
    packages=find_packages(),
    python_requires=">=3.7",
    include_package_data=True,
    tests_require=["pytest~=6.0.1"],
    install_requires=[
        "python-configuration~=0.7",
        "nltk~=3.5",
        "imageio~=2.6",
        "pillow~=7.1.0",
        "gputil~=1.4.0",
        "networkx~=2.4",
        "numpy~=1.18.5",
        "tensorflow-addons==0.11.1",
        "tensorflow==2.3.0",
    ],
    # metadata to display on PyPI
    author="Clemens-Alexander Brust",
    author_email="clemens-alexander.brust@uni-jena.de",
    description="Concept Hierarchies for Incremental and Active Learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 5 - Production/Stable",
        "Environment :: GPU",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    url="https://github.com/cvjena/chia",
)



from abc import ABC, abstractmethod


class IncrementalModel(ABC):
    @abstractmethod
    def observe(self, samples, gt_resource_id, progress_callback=None):
        return None

    @abstractmethod
    def predict(self, samples, prediction_resource_id):
        return None

    @abstractmethod
    def save(self, path):
        pass

    @abstractmethod
    def restore(self, path):
        pass


class ProbabilityOutputModel(IncrementalModel, ABC):
    def predict(self, samples, prediction_resource_id):
        samples_ = self.predict_probabilities(samples, prediction_resource_id + "_dist")
        samples_ = [
            sample.add_resource(
                self.__class__.__name__,
                prediction_resource_id,
                sorted(
                    sample.get_resource(prediction_resource_id + "_dist"),
                    key=lambda x: x[1],
                    reverse=True,
                )[0][0],
            )
            for sample in samples_
        ]
        return samples_

    @abstractmethod
    def predict_probabilities(self, samples, prediction_dist_resource_id):
        pass



import tensorflow as tf

from chia import components
from chia.components.base_models.keras import (
    keras_basemodel,
    keras_dataaugmentation,
    keras_featureextractor,
    keras_learningrateschedule,
    keras_preprocessor,
    keras_trainer,
)


def _get_input_img_np(sample):
    return sample.get_resource("input_img_np")


class KerasOptimizerFactory(components.Factory):
    name_to_class_mapping = {
        "adam": tf.keras.optimizers.Adam,
        "sgd": tf.keras.optimizers.SGD,
    }
    default_section = "keras_optimizer"
    i_know_that_var_args_are_not_supported = True


class KerasBaseModelContainer:
    def __init__(self, config, classifier, observers=()):
        self.learning_rate_schedule = keras_learningrateschedule.KerasLearningRateScheduleFactory.create(
            config["learning_rate_schedule"], observers=observers
        )
        self.optimizer = KerasOptimizerFactory.create(
            config["optimizer"], observers=observers
        )
        self.augmentation = keras_dataaugmentation.KerasDataAugmentationFactory.create(
            config["augmentation"], observers=observers
        )
        self.preprocessor = keras_preprocessor.KerasPreprocessorFactory.create(
            config["preprocessor"], observers=observers, augmentation=self.augmentation
        )

        self.feature_extractor = keras_featureextractor.KerasFeatureExtractorFactory.create(
            config["feature_extractor"],
            observers=observers,
            preprocessor=self.preprocessor,
        )

        self.trainer = keras_trainer.KerasTrainerFactory.create(
            config["trainer"],
            observers=observers,
            feature_extractor=self.feature_extractor,
            preprocessor=self.preprocessor,
            classifier=classifier,
            learning_rate_schedule=self.learning_rate_schedule,
            optimizer=self.optimizer,
        )

        self.base_model = keras_basemodel.KerasBaseModelFactory.create(
            dict(),  # This is because a field "trainer" is present in config -> conflict
            observers=observers,
            batch_size=config["trainer"]["batch_size"],
            classifier=classifier,
            feature_extractor=self.feature_extractor,
            trainer=self.trainer,
            preprocessor=self.preprocessor,
        )



import inspect
import typing

from chia import instrumentation


class Factory:
    name_to_class_mapping: typing.Optional[typing.Dict] = None
    default_section: typing.Optional[str] = None
    i_know_that_var_args_are_not_supported = False

    @classmethod
    def create(cls, config: dict, observers=(), **kwargs):
        temp_observable = instrumentation.NamedObservable(cls.__name__)
        for observer in observers:
            temp_observable.register(observer)

        if isinstance(cls.name_to_class_mapping, dict):
            name = config["name"]
            target_class = cls.name_to_class_mapping[name]
            temp_observable.notify(
                instrumentation.ConfigMessage(
                    cls.__name__, f"{cls.__name__}.name", name, source="config_dict",
                )
            )
        else:
            # If mapping is not a dict, interpret as type directly
            target_class = cls.name_to_class_mapping
            name = target_class.__name__

        init_method_signature = inspect.signature(target_class)

        call_spec_kwargs = dict()

        for parameter, param_spec in init_method_signature.parameters.items():
            # Sanity check
            if (
                param_spec.kind == inspect.Parameter.POSITIONAL_ONLY
                or param_spec.kind == inspect.Parameter.VAR_KEYWORD
                or param_spec.kind == inspect.Parameter.VAR_POSITIONAL
            ):
                if not cls.i_know_that_var_args_are_not_supported:
                    raise ValueError(
                        f"Unsupported kind of constructor parameter {parameter}"
                    )
                else:
                    # Skip the unsupported parameters
                    continue

            # Try to find it
            if parameter in kwargs.keys():
                param_value = kwargs[parameter]
            elif parameter in config.keys():
                param_value = config[parameter]
                temp_observable.notify(
                    instrumentation.ConfigMessage(
                        cls.__name__,
                        f"{target_class.__name__}.{parameter}",
                        param_value,
                        source="config_dict",
                    )
                )
            elif f"{name}_userdefaults.{parameter}" in config.keys():
                param_value = config[f"{name}_userdefaults.{parameter}"]
                temp_observable.notify(
                    instrumentation.ConfigMessage(
                        cls.__name__,
                        f"{target_class.__name__}.{parameter}",
                        param_value,
                        source="userdefaults",
                    )
                )
            elif param_spec.default != inspect.Signature.empty:
                param_value = param_spec.default
                temp_observable.notify(
                    instrumentation.ConfigMessage(
                        cls.__name__,
                        f"{target_class.__name__}.{parameter}",
                        param_value,
                        source="default",
                    )
                )

            else:
                raise ValueError(
                    f"Could not find a value for constructor parameter {parameter}"
                )

            call_spec_kwargs[parameter] = param_value

        # Call constructor
        instance = target_class(**call_spec_kwargs)

        # Register observers if possible
        if isinstance(instance, instrumentation.Observable):
            for observer in observers:
                instance.register(observer)

        for observer in observers:
            temp_observable.unregister(observer)

        return instance


class ContainerFactory(Factory):
    @classmethod
    def create(cls, config: dict, **kwargs):
        name = config["name"]
        target_class = cls.name_to_class_mapping[name]
        return target_class(config, **kwargs)
